#include "TwoLE.h"
#include "math.h"
#include "stdio.h"
#include "string.h"
#include "ctype.h"
#include "stdlib.h"

/**
 *   Classe permettant de decoder les TLE
 * 
TLE veut dire Two-Line Elements en anglais. Ces deux lignes continennent les
elements orbitaux qui définissent l'orbite exacte d'un satellite et la
position du satellite sur son orbite au moment precis appele epoque. L'epoque
des donnees orbitales, ou l'heure exacte a laquelle les donnees presentees sont
valides, est inscrite a la ligne un des TLE. Elle sert a determiner l'age des
elements, donc de savoir si les previsions qui seront generees avec eux seront
precises.
Les TLE d'un grand nombre de satellites sont accessibles sur le site :
http://celestrak.com/NORAD/elements/
 * @author Michel Llibre - ONERA/DCSD.
 */

// Suppression des espaces en fin de ligne
static char *trimfin(char *str)
{
    int n = strlen(str) -1;
    while (n > 0 && isspace(str[n])) {str[n] = 0; n--;}
    return str;
}

int TwoLE::flag_init = 0;
unsigned char TwoLE::val[256];

// Initialisation masque
void TwoLE::init()
{
    int i;
    for (i = 0  ; i <= 255; val[i] = 0, i++);
    for (i = '0'; i <= '9'; val[i] = (unsigned char) (i - '0'), i++);
    val['-'] = 1;
    flag_init = 1;
}

/**
 * Verification du format des TLE
 * @param line1
 * @param line2
 * @return 0 si OK, -1 si erreur Checksum, 1 si autre erreur
 */
int TwoLE::checkTLE(char *line1, char *line2)
{
    int x;
    int sum1, sum2;

    for (x=0, sum1=0, sum2=0; x<=67; sum1 +=  val[(int)line1[x]], sum2 +=  val[(int)line2[x]], x++);

    x = ( val[(int)line1[68]]^(sum1%10));
    x |= (val[(int)line2[68]]^(sum2%10));
    if (x != 0) return -1;
    x = (line1[0]^'1') ;
    x |= (line1[1]^' ') ;
    x |= (line1[7]^'U') ;
    x |= (line1[8]^' ') ;
    x |= (line1[17]^' ') ;
    x |= (line1[23]^'.') ;
    x |= (line1[32]^' ') ;
    x |= (line1[34]^'.') ;
    x |= (line1[43]^' ') ;
    x |= (line1[52]^' ') ;
    x |= (line1[61]^' ') ;
    x |= (line1[62]^'0') ;
    x |= (line1[63]^' ') ;
    x |= (line2[0]^'2')  ;
    x |= (line2[1]^' ')  ;
    x |= (line2[7]^' ')  ;
    x |= (line2[11]^'.') ;
    x |= (line2[16]^' ') ;
    x |= (line2[20]^'.') ;
    x |= (line2[25]^' ') ;
    x |= (line2[33]^' ') ;
    x |= (line2[37]^'.') ;
    x |= (line2[42]^' ') ;
    x |= (line2[46]^'.') ;
    x |= (line2[51]^' ') ;
    x |= (line2[54]^'.') ;
    x |= (line1[2]^line2[2]) ;
    x |= (line1[3]^line2[3]) ;
    x |= (line1[4]^line2[4]) ;
    x |= (line1[5]^line2[5]) ;
    x |= (line1[6]^line2[6]) ;
    x |= (isdigit(line1[68]) ? 0 : 1) ;
    x |= (isdigit(line2[68]) ? 0 : 1) ;
    x |= (isdigit(line1[18]) ? 0 : 1) ;
    x |= (isdigit(line1[19]) ? 0 : 1) ;
    x |= (isdigit(line2[31]) ? 0 : 1) ;
    x |= (isdigit(line2[32]) ? 0 : 1) ;

    return (x != 0 ? 1 : 0);
}

/**
 *  Construction et decodage d'une TLE à partir des 2 strings
 * @param name		String nom du satellite
 * @param line1		Ligne 1
 * @param line2		Ligne 2
 */
TwoLE::TwoLE(char * nom, char * line1, char * line2) {set(nom, line1, line2);}
void TwoLE::set(char * nom, char * line1, char * line2)
{
    char buf[16];
    if (!flag_init) init();
    double tempnum;
    if (name != NULL) strcpy(name, nom);
    designator[8] = 0; strncpy(designator, line1+9, 8); trimfin(designator);
    buf[ 5]=0; catnum = atoi(strncpy(buf, line1+2,5));
    buf[ 2]=0; year = atoi(strncpy(buf, line1+18,2));
    if (year <  56) year += 2000; else year += 1900;
    buf[12]=0; refepoch = strtod(strncpy(buf, line1+20, 12), NULL);
    buf[ 6]=0; tempnum = 1.0e-5*strtod(strncpy(buf, line1+44,6), NULL);
    nddot6 = tempnum/pow(10.0,(line1[51]-'0'));
    buf[ 6]=0; tempnum=1.0e-5*strtod(strncpy(buf, line1+53,6), NULL);
    bstar = tempnum/pow(10.0,(line1[60]-'0'));
    buf[ 4]=0; setnum = atoi(strncpy(buf, line1+64,4));
    buf[ 8]=0; incl = strtod(strncpy(buf, line2+8,8), NULL);
    buf[ 8]=0; raan = strtod(strncpy(buf, line2+17,8), NULL);
    buf[ 7]=0; eccn = 1.0e-07*strtod(strncpy(buf, line2+26,7), NULL);
    buf[ 8]=0; argper = strtod(strncpy(buf, line2+34,8), NULL);
    buf[ 8]=0; meanan = strtod(strncpy(buf, line2+43,8), NULL);
    buf[11]=0; meanmo = strtod(strncpy(buf, line2+52,11), NULL);
    buf[10]=0; drag = strtod(strncpy(buf, line1+33,10), NULL);
    buf[ 5]=0; orbitnum = atoi(strncpy(buf, line2+63,5));
}

/**
 * Construction et decodage d'une TwoLE lue dans un fichier
 * @param nom		Nom du satellite recherché
 * @param filename	Nom d'un fichier contenant des TLE
 * @return          La tle construite ou null si echec
 */
TwoLE::TwoLE(char *nom, char *filename)
{ if (!set(nom, filename)) fprintf(stderr,"Satellite %s non trouve dans le fichier %s\n", nom, filename); name[0] = 0;}

/* Renvoie 0 si pas trouvé */
int TwoLE::set(char *nom, char *filename)
{
    int nok = 0;
    char line1[71]; // mini 71 = 69 caractères + '\n' + '0' de fin
    char line2[71];
    FILE *in = fopen(filename,"rt");
    if (in == NULL)
    {
        fprintf(stderr, "Erreur acces fichier %s\n", filename);
        exit(1);
    }
    while(true)
    {
        int res, nn;
        if(NULL == fgets(name, 71, in)) break;
        trimfin(name);
        if(NULL == fgets(line1,71, in)) break;
        nn = strlen(line1);
        res = nn;
        if(NULL == fgets(line2,71, in)) break;
        if (0 != strcasecmp(nom, name)) continue;
        // break si lignes incompletes
        if (strlen(line1) < 69 || strlen(line2) < 69) break;
        res = checkTLE(line1,line2);
        if (res < 0)
            fprintf(stderr, "Erreur checksum TLE satellite %s\n", name);
        else if (res > 0)
            fprintf(stderr, "Erreur codage TLE satellite %s\n", name);
        else
        {
            set(name, line1, line2);
            nok = 1;
        }
        break;
    }
    fclose(in);
    return nok;
}

/**
 * Lecture de toutes les TLE d'un fichier et ajout des TLE lues au Vector tles
 * @return	le nombre de tle ajoutees
 */
int TwoLE::readFile(char * filename, vector<TwoLE *> *tles)
{
    TwoLE *tle;
    int n = 0;
    char name[71];
    char line1[71];
    char line2[71];
    FILE *in = fopen(filename,"rt");
    if (in == NULL)
    {
        fprintf(stderr, "Erreur acces fichier %s\n", filename);
        exit(1);
    }
        while(true)
        {
            if(NULL == fgets(name, 71, in)) break;
            if(NULL == fgets(line1,71, in)) break;
            if(NULL == fgets(line2,71, in)) break;

            trimfin(name);
            if (strlen(line1) < 69 || strlen(line2) < 69) break;
            int res = TwoLE::checkTLE(line1,line2);
            if (res < 0)
                fprintf(stderr, "Erreur checksum TLE satellite %s\n", name);
            else if (res > 0)
                fprintf(stderr, "Erreur codage TLE satellite %s\n", name);
            else
            {
                tle = new TwoLE(name, line1, line2);
                tles->push_back(tle);
                n++;
            }
        }
        fclose(in);
    return n;
}

/**
 * Impression d'un TwoLE
 */
void TwoLE::print()
{
    printf("Nom du satellite :                %s\n", name);
    printf("Catalog Number :                  %d\n", catnum);
    printf("Element set :                     %d\n", setnum);
    printf("Designation :                     %s\n", designator);
    printf("Annee de la date des donnees :    %d\n", year);
    printf("Jour dans l'annee :               %lf\n", refepoch);
    printf("Inclinaison orbite :              %lf degres\n", incl);
    printf("Ascension droite NA :             %lf degres\n", raan);
    printf("Excentricite :                    %lg\n", eccn);
    printf("Argument du perigee :             %lf degres\n", argper);
    printf("Anomalie moyenne a la date :      %lf degres\n", meanan);
    printf("Mouvement moyen :                 %lf  tours/jour\n", meanmo);
    printf("1/2 der1 mouv. moy :              %lg  tours/jour**2\n", drag);
    printf("1/6 der2 mouv. moy :              %lg  tours/jour**3\n", nddot6);
    printf("Trainee bstar :                   %lg ??\n", bstar);
    printf("Nombre de revolutions a la date : %d\n", orbitnum);
}

#include "satgeo.h"
void TwoLE::test()
{
    int res;
    printf("Début du Test : Construction SatGeo Spot vide, set à partir de spot.dat, impression Spot\n");

// Construction du satellite Spot à partir du fichier contenant ses éléments orbitaux
    SatGeo *Spot = new SatGeo();
    Spot->set("../CineSpace/espace/datas/sats/spot.dat");
    Spot->print();

// Construction du satellite Spot2 en extrayant ses éléments d'un fichier de TLE
    printf("\nConstruction tle, set tle Spot2 à partir resource.tle, impression tle\n");
    // Lecture des tle de Spot2
    TwoLE tle;
    res = tle.set("Spot 2", "../CineSpace/espace/datas/tle/resource.tle");
    if (res)
    {
        tle.print();
        printf("\nSet SatGeo Spot2 à partir tle Spot2, impression Spot2\n");
        SatGeo Spot2; Spot2.set(&tle);
        printf("\n");
        Spot2.print();
        double t = Spot2.t0;
        Spot2.etat(t, 1.e-6);
        printf("Etat standard\n"); Spot2.printEtat();
        Spot2.etatSGP(t, 1.e-6);
        printf("Etat SGP\n"); Spot2.printEtat();
    }
    else fprintf(stderr,"Pas trouvé Spot 2\n");

// Construction du satellite Spot4 en extrayant ses éléments d'un fichier de TLE
    printf("\nSet tle Spot4 à partir resource.tle\n");
    // Lecture des tle de Spot4
    res = tle.set("Spot 4", "../CineSpace/espace/datas/tle/resource.tle");
    if (res)
    {
        printf("\nSet SatGeo Spot4 à partir tle Spot4, impression Spot4\n");
        SatGeo Spot4; Spot4.set(&tle);
        printf("\n");
        Spot4.print();
    }
    else fprintf(stderr,"Pas trouvé Spot 4\n");

// Construction du satellite Spot5 en extrayant ses éléments d'un fichier de TLE
    printf("\nSet tle Spot5 à partir resource.tle\n");
    res = tle.set("Spot 5", "../CineSpace/espace/datas/tle/resource.tle");
    if (res)
    {
        printf("\nSet SatGeo Spot5 à partir tle Spot5, impression Spot5\n");
        SatGeo Spot5; Spot5.set(&tle);
        printf("\n");
        Spot5.print();
    }
    else fprintf(stderr,"Pas trouvé Spot 5\n");


// Construction du satellite Test1 en extrayant ses éléments d'un fichier de TLE
    printf("\nSet tle Test1 à partir test.tle\n");
    res = tle.set("Test1", "../CineSpace/espace/datas/tle/test.tle");
    if (res)
    {
        printf("\nSet SatGeo Test1 à partir tle Test1, impression Test1\n");
        SatGeo Test1; Test1.set(&tle);
        printf("\n");
        Test1.print();
    }
    else fprintf(stderr,"Pas trouvé Test1\n");

// Construction du satellite Test2 en extrayant ses éléments d'un fichier de TLE
    printf("\nSet tle Test2 à partir test.tle\n");
    res = tle.set("Test2", "../CineSpace/espace/datas/tle/test.tle");
    if (res)
    {
        printf("\nSet SatGeo Test2 à partir tle Test2, impression Test2\n");
        SatGeo Test2;
        Test2.set(&tle);
        printf("\n");
        Test2.print();
    }
    else fprintf(stderr,"Pas trouvé Test2\n");

// Construction d'un vector<TwoLE *> tle_res à partir du fichier resource.tle
    printf("\nConstruction vector<TwoLE *> tle_res. Lecture dans resource.tle\n");
    vector<TwoLE *> tle_res;
    TwoLE::readFile("../CineSpace/espace/datas/tle/resource.tle", &tle_res);

// Set SatGeo satRes3 à partir tle_res[3]
    SatGeo satRes3(tle_res[3]);
    printf("Set SatGeo satRes3 à partir tle_res[3], impression satRes3\n");
    printf("\n");
    satRes3.print();

    printf("\nSortie TwoLE::test()\n");
}
