#include <math.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "satgeo.h"
#include "terre.h"
#include "TwoLE.h"
#include "../mathutil/sp3.h"
#include "../mathutil/elem.h"

#include "../utils/cjulien/julien.h"

#define m00 0
#define m10 1
#define m20 2
#define m01 3
#define m11 4
#define m21 5
#define m02 6
#define m12 7
#define m22 8

/**
La classe SatGeo est une classe qui permet de definir les position, attitude 
d'un satellite et les vitesses associees a tout instant.
<P>
Dans la designation des composantes, les indices _n, _c, _g et _l sont relatifs aux
reperes suivants :
<UL>
<LI> Rn : repere lie a l'orbite avec x selon ligne des Noeuds et z selon H cinetique
<LI> Rc : repere equatorial Celeste (z vers le pole Nord et x vers Gamma J2000)
<LI> Rg : repere geographique (xOz contient Greenwich)
<LI> Rl : MON repere orbital Local (x~//Vit, y//Hcin, z//Terre->Sat )
<LI> Rs : Le repere lie au satellite
</UL>
<P>
Les unites pour toutes les variables sont en S.I. L'origine des temps (en secondes) est
le 1er janvier 2000 a midi a Greenwich (J2000.0).
@author Michel Llibre - ONERA/DCSD.
 */

void SatGeo::calFlagND()
{
    // Determination cote Noeud ascendant par rapport au Soleil
    double jj = t0/86400.;
    Terre::posSoleil(jj+jul2deltaET(jj + 2451545.), 1.e-6);
    double argNA = realmod2(gdOm0 - Terre::solAscen, 2.*M_PI);
    if (fabs(argNA) < M_PI/2) flag_ND = false;
    else flag_ND = true;
}


/** Construction d'un satellite a partir de ses elements orbitaux
 *  en unites S.I.
  @param nom	nom du satellite,
  @param n0	mouvement moyen(radians/s)
  @param e	excentricite
  @param i0	inclinaison de l'orbite (radians)
  @param om0	argument du perigee a la date t0 (radians)
  @param Om0	ascension droite du noeud ascendant a la date t0 (radians)
  @param M0	anomalie moyenne a la date t0 (radians)
  @param t0	date de mesure des elements orbitaux (secondes relatives J2000)
 */
SatGeo::SatGeo(char *name, double n0, double e, double i0, double om0, double Om0, double M0, double t0)
{ n0dots2 = 0.; n0ddots6 = 0.; set(name, n0, e, i0, om0, Om0, M0, t0);}
void SatGeo::set(char *name, double n0, double e, double i0, double om0, double Om0, double M0_, double t0_)
{
    if (name != NULL) strcpy(nom, name);
    OrbitGeo::set(n0, e, i0, om0, Om0);
    t0 = t0_;
    M0 = M0_;
    calFlagND();
    etat(t0, 1.e-6);
}


/** Creation d'un satellite a partir de ses elements orbitaux
 * lus dans un fichier. Les parametres lus sont dans l'ordre :
 * - le nom du satellite,
 * - la date t0 de mesure des elements orbitaux (lu en jours a partir de J2000)
 * - l'inclinaison i0 de l'orbite (lu en degres),
 * - l'ascension droite du noeud ascendant Om0 a la date t0 (lu en degres),
 * - l'excentricite,
 * - l'argument du perigee om0 a la date t0 (lu en degres),
 * - l'anomalie moyenne M0 a la date t0 (lue en degres),
 * - le mouvement moyen n0 (lu en tours/jour).
 * @param filename  Nom du fichier contenant les elements.
 */
SatGeo::SatGeo(char *filename) { if(!set(filename)) exit(1);}
// Renvoi 0 si echec
int SatGeo::set(char *filename)
{
char str[80], name[80];
double t0, i0, Om0, e0, om0, M0, n0;
int ln = 0, nok = 1;
FILE *fdata;

    fdata = fopen(filename, "rt");
    if (fdata == NULL)
    {
        fprintf(stderr, "Erreur acces fichier %s\n", filename);
        return 0;
    }
    for (;;)
    {
        ln++; fscanf(fdata,"%s %s", str, name);
        if (strcmp(str, "nom"))break;
        ln++; fscanf(fdata,"%s %lf", str, &t0);
        if(strcmp(str,"date_jours/J2000")) break; t0 *= 86400.;
        ln++; fscanf(fdata,"%s %lf", str, &i0);
        if(strcmp(str,"inclinaison_dg")) break; i0 *= M_PI/180.;
        ln++; fscanf(fdata,"%s %lf", str, &Om0);
        if(strcmp(str,"ascDroiteNA_dg")) break; Om0 *= M_PI/180.;
        ln++; fscanf(fdata,"%s %lf", str, &e0);
        if(strcmp(str,"excentricite")) break;
        ln++; fscanf(fdata,"%s %lf", str, &om0);
        if(strcmp(str,"argPerigee_dg")) break; om0 *= M_PI/180.;
        ln++; fscanf(fdata,"%s %lf", str, &M0);
        if(strcmp(str,"anomalieMoy_dg")) break; M0 *= M_PI/180.;
        ln++; fscanf(fdata,"%s %lf", str, &n0);
        if(strcmp(str,"mouvMoyen_tr/jour")) break; n0 *= 2.*M_PI/86400.;;
        nok = 0;
        break;
    }
    fclose(fdata);
    if (nok)
    {
        fprintf(stderr, "Erreur de lecture données fichier %s, ligne %d\n", filename, ln);
        return 0;
    }
    n0dots2 = 0.; n0ddots6 = 0.; set(name, n0, e0, i0, om0, Om0, M0, t0);
    return 1;
}

/** Construction d'un satellite a partir de ses elements orbitaux modifiés en S.I., de type CNES
    Dans la liste des arguements, le nom est placé en dernier pour avoir une signature différente
    du contruteucr de base.
         @param a   	demi grand axe (m)
         @param ex	excentricite*cos(omega)
         @param ey   	excentricite*sin(omega)
         @param incl  	inclinaison (radians)
         @param Om0  	ascension droite du noeud ascendant a la date t0 (radians)
         @param pso	argument du satellite a t0 (radians)
         @param t0  	date de mesure des elements orbitaux (secondes a partir de J2000)
         @param nom	nom du satellite,
         */
SatGeo::SatGeo(CnesParameters *p)
{ set(p->a, p->ex, p->ey, p->i0, p->gdOm0, p->pso, p->t0, p->nom);}
SatGeo::SatGeo(double a, double ex, double ey, double i0, double Om0, double pso, double t0, char * name)
{ set(a, ex, ey, i0, Om0, pso, t0, name);}
void SatGeo::set(double a, double ex, double ey, double i0, double Om0, double pso, double t0, char *name)
{
    double v, cos_v, sin_v, temp, cos_E, sin_E, E;
    e0 = sqrt(ex*ex+ey*ey);
    if (ey != 0. && ex != 0.) om0 = atan2(ey, ex);
    else om0 = 0.;
    // Correction J2
    // a /= OrbitGeo.corJ2(a, e, cos(i0));
    n0 = sqrt(Terre::GM/(a*a*a));
    v = pso - om0;
    cos_v = cos(v);
    sin_v = sin(v);
    temp = 1 + e0*cos_v;
    cos_E = (e0 + cos_v)/temp;
    sin_E = sqrt(1.-e0*e0)*sin_v/temp;
    E = atan2(sin_E, cos_E);
    //Calcul anomalie moyenne
    M0 = E - e0*sin_E;
    n0dots2 = 0.; n0ddots6 = 0.; set(name, n0, e0, i0, om0, Om0, M0, t0);
}


/**
 * Construction satellite a partir de ses TLE
 * @param tle
 */
SatGeo::SatGeo(TwoLE *tle)
{ set(tle);}
void SatGeo::set(TwoLE *tle)
{
    double coef1 = 2.*M_PI/Date_jour2s;
    double coef2 = coef1/Date_jour2s;
    double coef3 = coef2/Date_jour2s;
    double t1 = dat2jul(1., 1., tle->year);
    double t2 = tle->refepoch - 1.;
    double t3 = Date_J2000_j;
    t0 = (t1 + (t2)  - t3)*Date_jour2s;
    n0dots2 = tle->drag*coef2;
    n0ddots6 = tle->nddot6*coef3;
    set(tle->name, tle->meanmo*coef1, tle->eccn, toRadians(tle->incl), toRadians(tle->argper),
        toRadians(tle->raan), toRadians(tle->meanan), t0);
}

/** Creation d'un satellite a partir de sa position-vitesse :
    @param nom	nom du satellite,
    @param t0	date de mesure des elements (s a partir de J2000)
    @param v3d_TS_c	Vect3 Position satellite dans Rc
    @param v3d_Vs_c	Vect3 Vitesse satellite dans Rc
    @return new SatGeo cree.
	 */
SatGeo::SatGeo(char * name, double t0, double* v3d_TS_c, double* v3d_Vs_c)
{ if(!set(name, t0, v3d_TS_c, v3d_Vs_c)) exit(1);}
int SatGeo::set(char * name, double t0, double* v3d_TS_c, double*v3d_Vs_c)
{
    double i0, Om0, e, om0, n0, a;
    // Parametres locaux pour finaliser les parametres lus
    double V, dist, unsa, na2, CteAire, e2, sOmega = 0., cOmega = 1.;
    double calfa, salfa, cpo, spo;
    double sinE = 0., cosE = 1., E, unmecE, cosv, sinv;
    double ci0, si0, unme2;

    // Modules vitesse et distance
    V = longv3(v3d_Vs_c);
    dist = longv3(v3d_TS_c);
    if (dist == 0.)
    {
        fprintf(stderr, "Erreur constructeur d'orbite : Orbite non elliptique !");
        return 0;    }
    // Calcul grand axe par la relation de l'energie E = V2/2-mu/r = -mu/2a
    unsa = 2./dist - V*V/Terre::GM;
    if (unsa <= 0.)
    {
        fprintf(stderr, "Erreur constructeur d'orbite : Orbite non elliptique !");
        return 0;
    }
    a = 1./unsa;
    // Calcul moment cinetique et module
    crossv3(v3_hcin, v3d_TS_c, v3d_Vs_c);
    CteAire = longv3(v3_hcin);
    if (CteAire == 0.)
    {
        fprintf(stderr, "Erreur constructeur d'orbite : Orbite non elliptique !");
        return 0;
    }
    // Calcul vecteurs unitaires normal orbite et rayon vecteur
    scav3(v3_hcin, 1./CteAire, v3_hcin);
    scav3(v3_Zloc, 1./dist, v3d_TS_c);
    // Inclinaison
    ci0 = v3_hcin[2]; //forcement <= 1
    i0 = acos(ci0);
    si0 = sqrt(1.-ci0*ci0);
    // Ascension droite noeud ascendant
    if (si0 != 0.)
    {
        sOmega = v3_hcin[0]/si0;
        cOmega = -v3_hcin[1]/si0;
        calfa = v3_Zloc[0]*cOmega + v3_Zloc[1]*sOmega;
        salfa = v3_Zloc[2]/si0;
    }
    else // Orbite equatoriale
    {
        // sOmega = 0.; cOmega = 1.; deja fait
        calfa = v3_Zloc[0];
        salfa = v3_Zloc[1];
    }
    Om0 = atan2(sOmega ,cOmega);
    // Calcul excentricite et cie
    unme2 = CteAire*CteAire/(a*Terre::GM);
    e2 = 1. - unme2;
    if (unme2 <= 0. || e2 < 0.)
    {
        fprintf(stderr, "Erreur constructeur d'orbite : Orbite non elliptique !");
        return 0;
    }
    e = sqrt(e2);
    // Calcul n0 par 3eme loi de Kepler
    // a /= OrbitGeo.corJ2(a, e, ci0);
    na2 = sqrt(Terre::GM*a);
    n0 = na2/(a*a);
    // Calcul anomalies
    if (e > 0.)
    {
        sinE = dotv3(v3d_TS_c, v3d_Vs_c)/(e*na2);
        cosE = (1. - dist/a)/e;
        E = atan2(sinE, cosE);
        M0 = E - e*sinE;
        unmecE = 1. - e*cosE;
        cosv = (cosE - e)/unmecE;
        sinv = sqrt(unme2)*sinE/unmecE;
        cpo = calfa*cosv + salfa*sinv;
        spo = salfa*cosv - calfa*sinv;
        om0 = atan2(spo, cpo);
    }
    else
    {
        om0 = 0;
        M0 = atan2(salfa, calfa);
    }
    n0dots2 = 0.; n0ddots6 = 0.; set(name, n0, e0, i0, om0, Om0, M0, t0);
    return 1;
}


/** Creation d'une orbite heliosynchrone circulaire :
    @param nom   Nom du satellite
    @param Q     Nombre de jours de la periode de repetitivite
    @param K     Nombre de revolutions dans la periode de repetitivite
    @param SNA_h Ecart horaire decimal Soleil - noeud ascendant
    @param jNA   date de passage au noeud ascendant(jour decimal/J2000)
	 */
SatGeo::SatGeo(char * name, int Q, int K, double SNA_h, double jNA)
{ set(name, Q, K, SNA_h, jNA);}
void SatGeo::set(char * name, int Q, int K, double SNA_h, double jNA)
{
    double i0, Om0, n0;
    double nnod, OmP, ci0;
    double twoPI = 2.*M_PI;
    t0 = jNA*Date_jour2s;
    Terre::posSoleil(jNA+jul2deltaET(jNA + 2451545.), 1.e-6);
    Om0 = (SNA_h/24.-0.5)*twoPI + Terre::solAscen;
    OmP = twoPI/Terre::antrop2s;
    nnod = K*twoPI/(Q*Date_jour2s);
    // Calcul iteratif de x = Cos(i)
    double b1 = (2./3.)*(twoPI/Terre::antrop2s)/Terre::J2;
    double b2 = Terre::GM/(Terre::rT*Terre::rT*Terre::rT);
    double b = pow((b1*b1*b1*b2*b2), 1./7.);
    double sept_tiers = 7./3.;
    double xold, x = 0., y = 0.;
    int i;
    for (i = 0; i < 10; i++) // On limite a 10 iterations
    {
        xold = x;
        if (x != 0.) y = 0.5*OmP*(1. - 5.*x*x)/x;
        x = -pow(b/(nnod - y), sept_tiers);
        if (fabs(xold -x) <= 1.e-13) break;
    }
    if (x > 1.) x = 1.;
    else if (x < -1.) x = -1.;
    ci0 = x;
    i0 = acos(ci0);
    n0 = b/pow(-ci0, 3./7.);
    n0dots2 = 0.; n0ddots6 = 0.; set(name, n0, 0., i0, 0., Om0, 0., t0);
}


/** Mise en coherence des elements orbitaux d'un satellite a partir
 * d'un lot de ses parametres en unites S.I.
  @param n0		mouvement moyen(radians/s)
  @param e			excentricite
  @param i0		inclinaison de l'orbite (radians)
  @param om0		argument du perigee a la date t0 (radians)
  @param Om0		ascension droite du noeud ascendant a la date t0 (radians)
  @param M0		anomalie moyenne a la date t0 (radians)
  @param t0		date de mesure des elements orbitaux (secondes relatives J2000)
  @param tcou		date de calcul de l'etat (secondes relatives J2000)
 */
void SatGeo::update(double n0, double e, double i0,
                 double om0, double Om0, double M0_, double t0_, double tcou)
{
    OrbitGeo::set(n0, e, i0, om0, Om0);
    n0dots2 = 0.; n0ddots6 = 0.;
    t0 = t0_;
    M0 = M0_;
    calFlagND();
    etat(tcou, 1.e-6);  // tcou et pas t0 !!!!!
}

/** Calcul de l'etat satellite (cad de Ccn, TS_c, Vs_c, Ccl, Oml_c)
  @param t    date de calcul de l'etat en s a partir de J2000.0
 */
void SatGeo::etat(double t, double precision)
{
    if (tcou == t) return;
    tcou = t;
    double T, e, unme2, p, sq_unme2, sq_unme3s2;
    double dM, cE, sE, umecE, cpo, spo, cv, sv;
    double unpecosv, ct, st, temp, vP;
    int iter;
    T = t - t0;
    grand_Omega = gdOm0 + gdOmP*T;
    petit_omega = om0 + omP*T;
    //Mouvement moyen
    double n = n0 + T*(2.*n0dots2 + T*3.*n0ddots6);
    double a = a0 * pow(n0/n, 2./3.);
    e = 1.-a0*(1.-e0)/a;
    if (e < 0.) e = 0.;
    unme2 = 1-e*e;
    p = a*unme2;
    sq_unme2 = sqrt(unme2);
    sq_unme3s2 = unme2*sq_unme2;
    anoMoy_M = M0 + T*(n0 + T*(n0dots2 + T*n0ddots6));
    // Resolution equation de Kepler par iterations
    cpo = cos(petit_omega);
    spo = sin(petit_omega);
    double anoExc_E = anoMoy_M;
    sE = sin(anoExc_E);
    for (iter = 1; iter <= 4; iter++)
    {
        dM = anoMoy_M - (anoExc_E - e*sE);
        if (fabs(dM) < 1.e-13) break;
        anoExc_E += dM / (1. - e*cos(anoExc_E) ) ;
        sE = sin(anoExc_E);
    }
    cE = cos(anoExc_E); umecE = 1. - e*cE;
    cv = (cE - e)/umecE;
    sv = sq_unme2*sE/umecE;
    anoVraie_v = atan2(sv,cv);
    unpecosv = 1. + e*cv;
    dist = p/unpecosv;
    pso = petit_omega + anoVraie_v;
    ct = cos(pso); st = sin(pso);
    temp = a*n/sq_unme2;
    v3_Vs_n[0] = -temp*(st + e*spo);
    v3_Vs_n[1] = temp*(ct + e*cpo),
    v3_Vs_n[2] = 0.;
    // Attitude Rn / Rc
    double cgo, sgo;
    cgo = cos(grand_Omega);
    sgo = sin(grand_Omega);
    m3_CH[0] = cgo;  m3_CH[3] = -sgo; m3_CH[6] =  0.;
    m3_CH[1] = sgo;  m3_CH[4] =  cgo; m3_CH[7] =  0.;
    m3_CH[2] =  0.;  m3_CH[5] =   0.; m3_CH[8] =  1.;
    m3m3(m3_CN, m3_CH, m3_HN);
    // Attitude repere orbital local Rl / Rc
    // X selon Vit, Y selon Hcin, Zl selon TS
    m3_NL[0] = -st; m3_NL[3] =  0.; m3_NL[6] =  ct;
    m3_NL[1] =  ct; m3_NL[4] =  0.; m3_NL[7] =  st;
    m3_NL[2] =  0.; m3_NL[5] =  1.; m3_NL[8] =  0.;

    m3m3(m3_CL, m3_CN, m3_NL);
    // Vitesse rotation repere orbital local Rl / Rc
    vP = n*unpecosv*unpecosv/sq_unme3s2;
    temp = omP + vP;
    v3_Oml_c[0] = temp*m3_CN[m02];
    v3_Oml_c[1] = temp*m3_CN[m12];
    v3_Oml_c[2] = temp*m3_CN[m22] + gdOmP;
    ascen_droite = atan2(m3_CL[m12], m3_CL[m02]);
    declinaison = latitude = asin(m3_CL[m22]);
    Terre::orientation(t/86400., precision);
    tm3m3(m3_GL, Terre::m3_CG, m3_CL);
    longitude = atan2(m3_GL[m12], m3_GL[m02]);
    // Position et vitesses satellite dans Rc
    v3_TS_c[0] = dist*m3_CL[m02];
    v3_TS_c[1] = dist*m3_CL[m12];
    v3_TS_c[2] = dist*m3_CL[m22];
    m3v3(v3_Vs_c, m3_CN, v3_Vs_n);
}


/** Calcul de l'etat satellite (cad de Ccn, TS_c, Vs_c, Ccl, Oml_c)
 * Methode pour effectuer des test. Elle n'est pas validee.
 * Elle ne met pas à jour tcou de maniere a ne pas empecher le calcul
 * ulterieur par la methode etat(t) standard.
  @param t    date de calcul de l'etat en s a partir de J2000.0
 */
void SatGeo::etatSGP(double t, double precision)
{
    double dU;
    double ct, st, vP;
    int iter;
    double T = t - t0;
    double d1 = delta1;
    double a_0 = a0*(1.-d1*(1./3. + d1*(1. + 134.*d1/81.)));
    double p0 = a_0*(1.-e0*e0);
    double q0 = a_0*(1.-e0);
    double temp = Terre::r2J2*n0/(p0*p0);
    double Omp = -1.5*temp*ci0;
    double omp = 0.74*temp*(5.*ci0*ci0 - 1.);
    double n = n0 + T*(2.*n0dots2 + T*3.*n0ddots6);
    double a = a_0 * pow(n0/n, 2./3.);
    double e = 1.-q0/a;
    if (e < 1.e-7) e = 1.e-7;
    double unme2 = 1-e*e;
    double p = a*unme2;
    grand_Omega = gdOm0 + Omp*T;
    petit_omega = om0 + omp*T;
    anoMoy_M = M0 + T*(n0 + T*(n0dots2 + T*n0ddots6));
    double grand_Ls = anoMoy_M + grand_Omega + petit_omega;
    temp = Terre::J3*Terre::rT/(Terre::J2*p)*si0;
    double ay = e*sin(petit_omega) - 0.5*temp;
    double ax = e*cos(petit_omega);
    double grand_L = grand_Ls - 0.25*temp*ax*(3.+5.*ci0)/(1.+ci0);
    double U1 = grand_L - grand_Omega;
    double U = U1, cU=1., sU=0., ecE = 0., esE = 0.;
    double umecE = 1.;
    for (iter = 1; iter <= 4; iter++)
    {
        cU = cos(U); sU = sin(U);
        ecE = ax*cU + ay*sU;
        esE = ax*sU - ay*cU;
        umecE = 1.- ecE;
        dU = (U1 + esE - U)/umecE;
        if (fabs(dU) < 1.e-13) break;
        U = U + dU;
    }
    double eL2 = ax*ax + ay*ay;
    double umeL2 = 1. - eL2;
    double pL = a*umeL2;
    double r = a*umecE;
    temp = sqrt(Terre::GM)/r;
    double rprime = temp*sqrt(a)*esE;
    double rvprime = temp*sqrt(pL);
    temp = esE/(1.+sqrt(umeL2));
    double su = (sU-ay-ax*temp)/umecE;
    double cu = (cU - ax + ay*temp)/umecE;
    double u = atan2(su,cu);
    double c2u = cos(2.*u);
    double s2u = sin(2.*u);
    temp = Terre::r2J2/(pL*pL);
    double rk = r + 0.25*temp*pL*si0*si0*c2u;
    double uk = u - 0.125*temp*(7.*ci0*ci0-1.)*s2u;
    double grand_Omegak = grand_Omega + 0.75*temp*ci0*s2u;
    double ik = i0 + 0.75*temp*si0*ci0*c2u;
    pso = uk;
    dist = rk;
    anoVraie_v = pso - petit_omega;
    ct = cos(pso); st = sin(pso);
    v3_Vs_n[0] = rprime*ct-rvprime*st;
    v3_Vs_n[1] = rprime*st+rvprime*ct;
    v3_Vs_n[2] = 0.;
    // Attitude Rn / Rc
    double cgo, sgo;
    cgo = cos(grand_Omegak);
    sgo = sin(grand_Omegak);
    m3_CH[0] = cgo;  m3_CH[3] = -sgo; m3_CH[6] =  0.;
    m3_CH[1] = sgo;  m3_CH[4] =  cgo; m3_CH[7] =  0.;
    m3_CH[2] =  0.;  m3_CH[5] =   0.; m3_CH[8] =  1.;
    double ci = cos(ik);
    double si = sin(ik);
    m3_HN[0] = 1.;  m3_HN[3] = 0.; m3_HN[6] =  0.;
    m3_HN[1] = 0.;  m3_HN[4] = ci; m3_HN[7] = -si;
    m3_HN[2] = 0.;  m3_HN[5] = si; m3_HN[8] =  ci;

    m3m3(m3_CN, m3_CH, m3_HN);
    // Attitude repere orbital local Rl / Rc
    // X selon Vit, Y selon Hcin, Zl selon TS
    m3_NL[0] = -st; m3_NL[3] =  0.; m3_NL[6] =  ct;
    m3_NL[1] =  ct; m3_NL[4] =  0.; m3_NL[7] =  st;
    m3_NL[2] =  0.; m3_NL[5] =  1.; m3_NL[8] =  0.;

    m3m3(m3_CL, m3_CN, m3_NL);
    // Vitesse rotation repere orbital local Rl / Rc
    vP = n*sqrt(umeL2)/(umecE*umecE*(1.+ecE));
    temp = Omp + omp + vP;
    v3_Oml_c[0] =  temp*m3_CN[m02];
    v3_Oml_c[1] =  temp*m3_CN[m12];
    v3_Oml_c[2] =  temp*m3_CN[m22] + gdOmP;
    ascen_droite = atan2(m3_CL[m12], m3_CL[m02]);
    declinaison = latitude = asin(m3_CL[m22]);
    Terre::orientation(t/86400., precision);
    tm3m3(m3_GL, Terre::m3_CG, m3_CL);
    longitude = atan2(m3_GL[m12], m3_GL[m02]);
    // Position et vitesses satellite dans Rc
    v3_TS_c[0] = dist*m3_CL[m02];
    v3_TS_c[1] = dist*m3_CL[m12];
    v3_TS_c[2] = dist*m3_CL[m22];
    m3v3(v3_Vs_c, m3_CN, v3_Vs_n);
}
#if 0
/**
	 *  Calcul de l'instant correspondant a une position sur orbite donnee
	 *  dans la revolution qui suit un instant donne
	 *  
	 * @param tref  instant origine du calcul
	 * @param alpha position sur orbite (angle en radian / noeud Ascendant)
	 * @return      instant de passage à l'angle alpha
	 */
double SatGeo::pso2time(double tref, double alpha)
{
    double Mref = M0 + n0*(tref - t0);// Anomalie moyenne a t = tref
    //Calcul anomalie vraie
    double v = alpha - petit_omega;
    //Calcul anomalie excentrique
    double cos_v = cos(v);
    double sin_v = sin(v);
    double temp = 1 + e0*cos_v;
    double cos_E = (e0 + cos_v)/temp;
    double sin_E = sin_v*sqrt(1.-e0*e0)/temp;
    double E = atan2(sin_E, cos_E);
    //Calcul anomalie moyenne
    double M = E - e0*sin_E;
    //Calcul instant correspondant
    return tref + Mlu.mod(M - Mref, 2.*M_PI)/n0;
}
#endif
/**
 * Impression paramètres orbitaux
 */
void SatGeo::print()
{
    printf("\nParametres orbitaux du satellite %s  dans Rc\n", nom);
    printf("Date t0 des par.   : %lf  s/J2000\n", t0);
    printf("Anomalie moy. à t0 : %lf dg\n", toDegrees(M0));
    OrbitGeo::print();
    if (flag_ND) printf("Le noeud ascendant est cote nuit\n");
    else printf("Le noeud ascendant est cote jour\n");
}


/**
 * Impression de l'etat du satellite
 */
void SatGeo::printEtat()
{
    char ligne[80];
    double W[3];

    printf("\nEtat du satellite %s dans Rc le %s\n", nom, jul2str(ligne, tcou/Date_jour2s + Date_J2000_j));
    printf("Distance Terre-Sat : %lf km\n", dist/1000.);
    printf("Anomalie moy.      : %lf dg\n", toDegrees(anoMoy_M));
    printf("Anomalie vraie.    : %lf dg\n", toDegrees(anoVraie_v));
    printf("Position sur orbite: %lf dg\n", toDegrees(pso));
    scav3(W, 1.e-3, v3_TS_c);
    printf("Position/Rc        : %s km\n", sprintv3(ligne, W, 10, 3));
    printf("Ascension droite   : %lf dg\n", toDegrees(ascen_droite));
    printf("Longitude          : %lf dg\n", toDegrees(longitude));
    printf("Latitude           : %lf dg\n", toDegrees(declinaison));
    printf("Vitesse trans./ Rc : %s m/s\n", sprintv3(ligne, v3_Vs_c, 10, 3));
    scav3(W, Date_jour2s*0.5/M_PI ,v3_Oml_c);
    printf("Vit. rot. dans Rc  : %s tr/j\n", sprintv3(ligne, W, 7, 3));
}

// Procédure statique de test
void SatGeo::test()
{
    double W[3];
    char ligne[80];
    SatGeo Spot;
    Spot.set("../CineSpace/espace/datas/sats/spot.dat");
    double t = Spot.t0;
    printf("Date t0 : %lf jours/J2000\n", Spot.t0/Date_jour2s);
    Spot.print();
    Spot.etat(t, 1.e-6);
    Spot.printEtat();
    double SNA_h = 12*(Spot.gdOm0 - Terre::solAscen)/M_PI + 12;
    double jNA = (Spot.t0 - Spot.M0/Spot.n0)/Date_jour2s;
    printf("Arg. N.A/Sol       : %lf heures\n", SNA_h);
    printf("Passage au N.A.    : %lf jours/J2000\n", jNA);

    SatGeo Spot2;
    Spot2.set("newSpot", t, Spot.v3_TS_c, Spot.v3_Vs_c);
    Spot2.print();
    Spot2.etat(t, 1.e-6);
    Spot2.printEtat();
    printf("\nAu perigee :\n");
    printf("Position Spot      : %s\n", sprintv3(ligne, Spot.v3_TS_c, 10,1));
    printf("Position Spot2     : %s\n", sprintv3(ligne, Spot2.v3_TS_c, 10, 1));
    subv3(W, Spot.v3_TS_c, Spot2.v3_TS_c);
    printf("Ecart de position  : %lf\n", longv3(W));
    printf("Vitesse Spot       : %s\n", sprintv3(ligne, Spot.v3_Vs_c, 7,1));
    printf("Vitesse Spot2      : %s\n", sprintv3(ligne, Spot2.v3_Vs_c, 7,1));
    subv3(W, Spot.v3_Vs_c, Spot2.v3_Vs_c);
    printf("Ecart de Vitesse   : %lf\n", longv3(W));
    printf("\nA l'apogee : \n");
    t = t + 0.5*M_PI/Spot.n0;
    Spot.etat(t, 1.e-6);
    Spot2.etat(t, 1.e-6);
    printf("Position Spot      : %s\n", sprintv3(ligne, Spot.v3_TS_c, 10,1));
    printf("Position Spot2     : %s\n", sprintv3(ligne, Spot2.v3_TS_c, 10, 1));
    subv3(W, Spot.v3_TS_c, Spot2.v3_TS_c);
    printf("Ecart de position  : %lf\n", longv3(W));
    printf("Vitesse Spot       : %s\n", sprintv3(ligne, Spot.v3_Vs_c, 7,1));
    printf("Vitesse Spot2      : %s\n", sprintv3(ligne, Spot2.v3_Vs_c, 7,1));
    subv3(W, Spot.v3_Vs_c, Spot2.v3_Vs_c);
    printf("Ecart de Vitesse   : %lf\n", longv3(W));
    t = t + 0.5*M_PI/Spot.n0;
    Spot.etat(t, 1.e-6);
    Spot2.etat(t, 1.e-6);
    printf("Position Spot      : %s\n", sprintv3(ligne, Spot.v3_TS_c, 10,1));
    printf("Position Spot2     : %s\n", sprintv3(ligne, Spot2.v3_TS_c, 10, 1));
    subv3(W, Spot.v3_TS_c, Spot2.v3_TS_c);
    printf("Ecart de position  : %lf\n", longv3(W));
    printf("Vitesse Spot       : %s\n", sprintv3(ligne, Spot.v3_Vs_c, 7,1));
    printf("Vitesse Spot2      : %s\n", sprintv3(ligne, Spot2.v3_Vs_c, 7,1));
    subv3(W, Spot.v3_Vs_c, Spot2.v3_Vs_c);
    printf("Ecart de Vitesse   : %lf\n", longv3(W));
    t = t + 0.5*M_PI/Spot.n0;
    Spot.etat(t, 1.e-6);
    Spot2.etat(t, 1.e-6);
    printf("Position Spot      : %s\n", sprintv3(ligne, Spot.v3_TS_c, 10,1));
    printf("Position Spot2     : %s\n", sprintv3(ligne, Spot2.v3_TS_c, 10, 1));
    subv3(W, Spot.v3_TS_c, Spot2.v3_TS_c);
    printf("Ecart de position  : %lf\n", longv3(W));
    printf("Vitesse Spot       : %s\n", sprintv3(ligne, Spot.v3_Vs_c, 7,1));
    printf("Vitesse Spot2      : %s\n", sprintv3(ligne, Spot2.v3_Vs_c, 7,1));
    subv3(W, Spot.v3_Vs_c, Spot2.v3_Vs_c);
    printf("Ecart de Vitesse   : %lf\n", longv3(W));
    t = t + 0.5*M_PI/Spot.n0;
    Spot.etat(t, 1.e-6);
    Spot2.etat(t, 1.e-6);
    printf("Position Spot      : %s\n", sprintv3(ligne, Spot.v3_TS_c, 10,1));
    printf("Position Spot2     : %s\n", sprintv3(ligne, Spot2.v3_TS_c, 10, 1));
    subv3(W, Spot.v3_TS_c, Spot2.v3_TS_c);
    printf("Ecart de position  : %lf\n", longv3(W));
    printf("Vitesse Spot       : %s\n", sprintv3(ligne, Spot.v3_Vs_c, 7,1));
    printf("Vitesse Spot2      : %s\n", sprintv3(ligne, Spot2.v3_Vs_c, 7,1));
    subv3(W, Spot.v3_Vs_c, Spot2.v3_Vs_c);
    printf("Ecart de Vitesse   : %lf\n", longv3(W));

    t = Spot.t0 +Spot.Tn0/3.;
    double om = Spot.om0 + Spot.omP*(t-Spot.t0);
    double Om = Spot.gdOm0 + Spot.gdOmP*(t-Spot.t0);
    Spot.etat(t, 1.e-6);
    SatGeo Spot3;
    Spot3.set(Spot.a0, Spot.e0*cos(om), Spot.e0*sin(om), Spot.i0, Om, Spot.pso, t, "Spot3");
    Spot3.etat(t, 1.e-6);
    subv3(W, Spot.v3_TS_c, Spot3.v3_TS_c);
    printf("Ecart de position 0-3  : %lf\n", longv3(W));
    t = t + Spot.Tn0/2.;
    subv3(W, Spot.v3_TS_c, Spot3.v3_TS_c);
    printf("Ecart de position 0-3  : %lf\n", longv3(W));

    t = t+10000.;
    Spot.etat(t, 1.e-6);
    Spot3.etat(t, 1.e-6);
    subv3(W, Spot.v3_TS_c, Spot3.v3_TS_c);
    printf("Ecart de position 0-3  : %lf\n", longv3(W));
    t = t + Spot.Tn0/2.;
    subv3(W, Spot.v3_TS_c, Spot3.v3_TS_c);
    printf("Ecart de position 0-3  : %lf\n", longv3(W));
    Spot3.printEtat();

    SNA_h = 12*(Spot3.gdOm0 - Terre::solAscen)/M_PI + 12;
    jNA = (Spot3.t0 - Spot3.M0/Spot3.n0)/Date_jour2s;
    SatGeo Spot4; Spot4.set("Spot4", Spot3.Qrep, Spot3.Krep, SNA_h, jNA);
    Spot4.etat(t, 1.e-6);
    Spot4.printEtat();
}
